# -*- coding: utf-8 -*-

"""
[CN] maintainer guide

这个模块中定义的是 ``aws_console_url_search/data.json`` 中的数据结构.
"""

import typing as T
import json
import dataclasses

from .constants import MAX_SERVICE_RANK, MAX_MENU_RANK
from .paths import path_data_json


@dataclasses.dataclass
class BaseModel:
    @classmethod
    def from_dict(cls, data: T.Dict[str, T.Any]):
        return cls(**data)

    def to_dict(self) -> T.Dict[str, T.Any]:
        return dataclasses.asdict(self)


@dataclasses.dataclass
class Service:
    # fmt: off
    id: str = dataclasses.field()
    name: str = dataclasses.field()
    url: str = dataclasses.field()
    description: str = dataclasses.field()
    globally: bool = dataclasses.field()  # global is a python reserved keyword, so we have to use globally
    terms: T.Optional[str] = dataclasses.field()
    emoji: T.Optional[str] = dataclasses.field()
    rank: T.Optional[int] = dataclasses.field()
    menus: T.List["Menu"] = dataclasses.field(default_factory=list)
    # fmt: on

    @property
    def sort_key(self) -> str:
        return f"{str(self.rank).zfill(8)}-{str(self.id).zfill(50)}"


@dataclasses.dataclass
class Menu:
    id: str = dataclasses.field()
    name: str = dataclasses.field()
    url: str = dataclasses.field()
    description: str = dataclasses.field()
    terms: T.Optional[str] = dataclasses.field()
    rank: int = dataclasses.field()

    @property
    def sort_key(self) -> str:
        return f"{str(self.rank).zfill(8)}-{str(self.id).zfill(50)}"


def load_data() -> T.List[Service]:
    console_url_data = json.loads(path_data_json.read_text())
    service_list = list()
    for service_id, service_data in console_url_data.items():
        menus = list()
        for menu_id, menu_data in service_data.get("menus", {}).items():
            menu = Menu(
                id=menu_id,
                name=menu_data["name"],
                url=menu_data["url"],
                description=menu_data.get("description", menu_data["name"]),
                terms=menu_data.get("terms"),
                rank=menu_data.get("rank", MAX_MENU_RANK),
            )
            menus.append(menu)
        service = Service(
            id=service_id,
            name=service_data["name"],
            url=service_data["url"],
            description=service_data.get("description", service_data["name"]),
            globally=service_data.get("global", False),
            terms=service_data.get("terms"),
            emoji=service_data.get("emoji"),
            rank=service_data.get("rank", MAX_SERVICE_RANK),
            menus=menus,
        )
        service_list.append(service)
    return service_list



# -*- coding: utf-8 -*-

"""
[CN] maintainer guide

这个模块定义了所有要用到的 ``sayt.DataSet``.
"""

import typing as T
import csv
import dataclasses

import sayt.api as sayt

from .paths import (
    dir_index,
    dir_cache,
    path_region_json,
)
from .model import Service, Menu, load_data
from .constants import MAX_SERVICE_RANK, MAX_MENU_RANK


def preprocess_query(query: T.Optional[str]) -> str:  # pragma: no cover
    """
    Preprocess query, automatically add fuzzy search term if applicable.
    """
    delimiter = ".-_@+"
    if query:
        for char in delimiter:
            query = query.replace(char, " ")
        words = list()
        for word in query.split():
            if word.strip():
                word = word.strip()
                if len(word) == 1:
                    if word == "*":
                        words.append(word)
                else:
                    # for fuzzy search, the first two characters must be matched
                    try:
                        if word[-2] != "~" and not word.endswith("!~"):
                            word = f"{word}~1/2"
                    except IndexError:
                        word = f"{word}~1/2"
                    words.append(word)
        if words:
            return " ".join(words)
        else:
            return "*"
    else:
        return "*"


# ------------------------------------------------------------------------------
# AWS Service and Menu
# ------------------------------------------------------------------------------
# fmt: off
fields = [
    sayt.IdField(name="id", stored=True, field_boost=10.0),
    sayt.TextField(name="id_text", stored=True, field_boost=10.0),
    sayt.StoredField(name="srv_name"),
    sayt.StoredField(name="menu_name"),
    sayt.StoredField(name="url"),
    sayt.StoredField(name="desc"),
    sayt.StoredField(name="globally"), # global is a python reserved keyword, so we have to use globally
    sayt.StoredField(name="emoji"),
    sayt.TextField(name="srv_text", stored=True, field_boost=7.5),
    sayt.NgramWordsField(name="srv_ngram", stored=True, minsize=2, maxsize=6, field_boost=5.0),
    sayt.TextField(name="menu_text", stored=True, field_boost=2.0),
    sayt.NgramWordsField(name="menu_ngram", stored=True, minsize=2, maxsize=6, field_boost=1.0),
    sayt.NumericField(name="rank", stored=True, sortable=True, ascending=True),
]
# fmt: on


@dataclasses.dataclass
class ServiceDocument:
    # fmt: off
    id: str = dataclasses.field()
    id_text: str = dataclasses.field()
    srv_name: str = dataclasses.field()
    menu_name: T.Optional[str] = dataclasses.field()
    url: str = dataclasses.field()
    desc: str = dataclasses.field()
    globally: bool = dataclasses.field() # global is a python reserved keyword, so we have to use globally
    emoji: T.Optional[str] = dataclasses.field()
    rank: int = dataclasses.field()
    srv_text: str = dataclasses.field()
    srv_ngram: str = dataclasses.field()
    menu_text: T.Optional[str] = dataclasses.field()
    menu_ngram: T.Optional[str] = dataclasses.field()
    # fmt: on

    @classmethod
    def from_service(cls, service: Service):
        service_rank = service.rank if service.rank else MAX_SERVICE_RANK
        if service.terms:
            srv_text = f"{service.name} {service.terms}"
        else:
            srv_text = service.name
        return cls(
            id=service.id,
            id_text=service.id,
            srv_name=service.name,
            menu_name=None,
            url=service.url,
            desc=service.description,
            globally=service.globally,
            emoji=service.emoji,
            rank=0 - (MAX_SERVICE_RANK - service_rank),
            srv_text=srv_text,
            srv_ngram=srv_text,
            menu_text=None,
            menu_ngram=None,
        )

    @classmethod
    def from_menu(cls, service: Service, menu: Menu):
        service_rank = service.rank if service.rank else MAX_SERVICE_RANK
        menu_rank = menu.rank if menu.rank else MAX_MENU_RANK
        if service.terms:
            srv_text = f"{service.name} {service.terms}"
        else:
            srv_text = service.name
        if menu.terms:
            menu_text = f"{menu.name} {menu.terms}"
        else:
            menu_text = menu.name
        return cls(
            id=menu.id,
            id_text=menu.id,
            srv_name=service.name,
            menu_name=menu.name,
            url=menu.url,
            desc=menu.description,
            globally=service.globally,
            emoji=service.emoji,
            rank=service_rank * MAX_MENU_RANK + menu_rank,
            srv_text=srv_text,
            srv_ngram=srv_text,
            menu_text=menu_text,
            menu_ngram=menu_text,
        )

    @classmethod
    def from_result(cls, doc: sayt.T_DOCUMENT):
        doc.setdefault("menu_name", None)
        doc.setdefault("emoji", None)
        doc.setdefault("menu_text", None)
        doc.setdefault("menu_ngram", None)
        return cls(**doc)


def service_downloader() -> T.List[sayt.T_DOCUMENT]:
    """
    Read data from ``path_data_json`` file.
    """
    docs = list()
    service_list = load_data()
    for service in service_list:
        docs.append(ServiceDocument.from_service(service))
        for menu in service.menus:
            docs.append(ServiceDocument.from_menu(service, menu))
    return [dataclasses.asdict(doc) for doc in docs]


service_index_name = "services"
service_dataset = sayt.DataSet(
    dir_index=dir_index,
    index_name=service_index_name,
    fields=fields,
    dir_cache=dir_cache,
    cache_key=service_index_name,
    cache_tag=service_index_name,
    cache_expire=24 * 60 * 60,
    downloader=service_downloader,
)


# ------------------------------------------------------------------------------
# AWS Region
# ------------------------------------------------------------------------------
@dataclasses.dataclass
class RegionDocument:
    id: str = dataclasses.field()
    region: str = dataclasses.field()
    desc: str = dataclasses.field()


def region_downloader():
    docs = list()
    with path_region_json.open("r", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        for region, desc in reader:
            docs.append(dict(id=region, region=region, desc=desc))
    return docs


region_index_name = "regions"
region_dataset = sayt.DataSet(
    dir_index=dir_index,
    index_name=region_index_name,
    fields=[
        sayt.IdField(name="id", stored=True),
        sayt.NgramWordsField(name="region", stored=True, minsize=2, maxsize=6),
        sayt.StoredField(name="desc"),
    ],
    dir_cache=dir_cache,
    cache_key=region_index_name,
    cache_tag=region_index_name,
    cache_expire=60,
    downloader=region_downloader,
)



# -*- coding: utf-8 -*-

import enum

MAX_SERVICE_RANK = 10000
MAX_MENU_RANK = 1000


class _AwsConsoleDomain:
    domain = "amazon.com"
    subdomain = "console.aws"


class _AwsUsGovConsoleDomain:
    domain = "amazonaws-us-gov.com"
    subdomain = "console"


class _AwsCNConsoleDomain:
    domain = "amazonaws-cn.com"
    subdomain = "console"


class AwsConsoleDomain(enum.Enum):
    """
    This is the full domain (subdomain + domain)

    Reference:

    - https://docs.aws.amazon.com/govcloud-us/latest/UserGuide/signing-into-govcloud.html
    """

    aws = _AwsConsoleDomain
    aws_us_gov = _AwsUsGovConsoleDomain
    aws_cn = _AwsCNConsoleDomain



# -*- coding: utf-8 -*-

"""
A cross-platform CLI tool for lightning-fast AWS Console URL opener.
"""


from ._version import __version__

__short_description__ = "A cross-platform CLI tool for lightning-fast AWS Console URL opener."
__license__ = "GPL"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"



# -*- coding: utf-8 -*-

import aws_console_url_search.api as api


def test():
    _ = api


if __name__ == "__main__":
    from aws_console_url_search.tests import run_cov_test

    run_cov_test(__file__, "aws_console_url_search.api", preview=False)



# -*- coding: utf-8 -*-

"""
todo: docstring
"""

import typing as T
import fire

from .._version import __version__


class AcsCli:
    """
    AWS Console Url Search CLI.
    """

    def __call__(self, version: T.Optional[bool] = None):
        """
        Serve for the ``acs`` command without any arguments.
        """
        from ..ui_init import ui, run_ui

        if version:
            print(__version__)
        else:
            run_ui()


def run():
    """
    The entry point of this CLI tool.
    """
    ars_cli = AcsCli()
    fire.Fire(ars_cli)



# -*- coding: utf-8 -*-

"""
todo: docstring
"""

import typing as T
import dataclasses

import zelfred.api as zf

from .dataset import ServiceDocument, service_dataset, preprocess_query, region_dataset
from .paths import path_current_region

# ------------------------------------------------------------------------------
# Item
# ------------------------------------------------------------------------------
@dataclasses.dataclass
class Item(zf.Item):
    def post_enter_handler(self, ui: "UI"):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_a_handler(self, ui: "UI"):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_w_handler(self, ui: "UI"):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_u_handler(self, ui: "UI"):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_p_handler(self, ui: "UI"):  # pragma: no cover
        ui.wait_next_user_input()


@dataclasses.dataclass
class InfoItem(Item):
    pass


# ------------------------------------------------------------------------------
# search_service_handler
# ------------------------------------------------------------------------------
@dataclasses.dataclass
class ConsoleUrlItem(Item):
    @classmethod
    def from_doc(
        cls,
        ui: "UI",
        doc: ServiceDocument,
        console_domain: str,
        aws_region: T.Optional[str],
    ):
        service_id = doc.id.split("-", 1)[0]
        if doc.emoji:
            emoji = f"{doc.emoji} "
        else:
            emoji = ""
        if doc.menu_name:
            title = f"🌠 {emoji}{ui.format_resource_type(service_id)} | {doc.menu_name}"
        else:
            title = f"🌟 {emoji}{ui.format_resource_type(service_id)}"
        if aws_region:
            region_part = f"region={aws_region}"
        else:
            region_part = ""
        if doc.url.startswith("http"):
            url = doc.url
        else:
            url = "https://{}{}".format(
                console_domain, doc.url.format(region=region_part)
            )
        return cls(
            title=title,
            subtitle=f"{doc.desc}",
            uid=doc.id,
            autocomplete=doc.id,
            variables={"doc": doc, "url": url},
        )

    def enter_handler(self, ui: "UI"):  # pragma: no cover
        """
        Open the url in browser. If not working, then print the url.
        """
        zf.open_url_or_print(self.variables["url"])

    def ctrl_a_handler(self, ui: "UI"):  # pragma: no cover
        """
        Copy the URL to clipboard. If not working, then print the url.
        """
        self.ctrl_u_handler(ui)

    def ctrl_u_handler(self, ui: "UI"):  # pragma: no cover
        """
        Copy the URL to clipboard. If not working, then print the url.
        """
        zf.copy_or_print(self.variables["url"])


def search_service_and_return_items(
    query: str,
    ui: T.Optional["UI"] = None,
    refresh_data: bool = False,
) -> T.List[T.Union["ConsoleUrlItem", "InfoItem"]]:
    query = preprocess_query(query)
    docs = service_dataset.search(
        query=query,
        limit=50,
        simple_response=True,
        verbose=False,
        refresh_data=refresh_data,
    )
    if len(docs):
        docs = [ServiceDocument.from_result(doc) for doc in docs]
        console_domain = ui.console_domain
        return [
            ConsoleUrlItem.from_doc(
                ui=ui,
                doc=doc,
                console_domain=console_domain,
                aws_region=ui.aws_region,
            )
            for doc in docs
        ]
    else:
        return [
            InfoItem(
                title="No service found",
                subtitle="",
                uid="no-service-found",
            )
        ]


def search_service_handler(
    query: str,
    ui: T.Optional["UI"] = None,
    skip_ui: bool = False,
):
    if skip_ui is False:  # pragma: no cover
        ui.render.prompt = f"[Query (region = {ui.aws_region})]"
    if query.strip().endswith("!~"):
        if skip_ui is False:  # pragma: no cover
            ui.run_handler(items=[])
            ui.repaint()
            ui.line_editor.press_backspace(n=2)
        return search_service_and_return_items(
            ui=ui,
            query=query[:-2],
            refresh_data=True,
        )

    items = search_service_and_return_items(
        ui=ui,
        query=query,
        refresh_data=False,
    )
    return items


# ------------------------------------------------------------------------------
# search_region_handler
# ------------------------------------------------------------------------------
@dataclasses.dataclass
class RegionItem(Item):
    @classmethod
    def from_region(
        cls,
        ui: "UI",
        line_input: str,
        region: str,
        description: str,
    ):
        return cls(
            title=f"{ui.format_highlight(region)} | {description}",
            subtitle=f"Hit {ui.ENTER} set region and return to search",
            uid=region,
            autocomplete=line_input + "!@" + region,
            variables={"region": region, "line_input": line_input},
        )

    def enter_handler(self, ui: "UI"):
        """
        Update ``UI.aws_region`` attribute and update the
        ``${HOME}/.aws_console_url_search/current_region.txt`` file.
        """
        region = self.variables["region"]
        if region == "no region":
            ui.aws_region = None
            path_current_region.write_text("")
        else:
            ui.aws_region = region
            path_current_region.write_text(region)

    def post_enter_handler(self, ui: "UI"):  # pragma: no cover
        """
        When exiting the switch aws region session, recover the original query input.
        """
        ui.line_editor.clear_line()
        ui.line_editor.enter_text(self.variables["line_input"])

    def ctrl_a_handler(self, ui: "UI"):  # pragma: no cover
        """
        Copy the AWS region to clipboard. If not working, then print the region.
        """
        zf.copy_or_print(self.variables["region"])

    def post_ctrl_a_handler(self, ui: "UI"):  # pragma: no cover
        self.post_enter_handler(ui)

    def ctrl_u_handler(self, ui: "UI"):  # pragma: no cover
        """
        Copy the AWS region to clipboard. If not working, then print the region.
        """
        self.ctrl_a_handler(ui)

    def post_ctrl_u_handler(self, ui: "UI"):  # pragma: no cover
        self.post_enter_handler(ui)


def search_region_and_return_items(
    line_input: str,
    aws_region_query: str,
    ui: T.Optional["UI"] = None,
    refresh_data: bool = True,
) -> T.List[T.Union["RegionItem", "InfoItem"]]:
    query = preprocess_query(aws_region_query)
    docs = region_dataset.search(
        query=query,
        limit=50,
        simple_response=True,
        verbose=False,
        refresh_data=refresh_data,
    )
    if len(docs):
        return [
            RegionItem.from_region(
                ui=ui,
                line_input=line_input,
                region=doc["region"],
                description=doc["desc"],
            )
            for doc in docs
        ]
    else:
        return [
            InfoItem(
                title="No region found",
                subtitle=f"Hit {ui.TAB} to return to re-enter a region query",
                autocomplete=line_input + "!@",
                uid="no-region-found",
            )
        ]


def search_region_handler(
    line_input: str,
    aws_region_query: str,
    ui: T.Optional["UI"] = None,
    skip_ui: bool = False,
):
    if skip_ui is False:  # pragma: no cover
        ui.render.prompt = f"(AWS Region Query)"
    return search_region_and_return_items(
        ui=ui,
        line_input=line_input,
        aws_region_query=aws_region_query,
    )


# ------------------------------------------------------------------------------
# main handler
# ------------------------------------------------------------------------------
def handler(
    query: str,
    ui: T.Optional["UI"] = None,
    skip_ui: bool = False,
) -> T.List[T.Union["ConsoleUrlItem", "RegionItem", "InfoItem"]]:
    """
    The main UI handler.
    """
    # user can use !@ to switch aws region
    if len(query.split("!@", 1)) > 1:
        line_input, aws_region_query = query.split("!@", 1)
        return search_region_handler(
            ui=ui,
            line_input=line_input,
            aws_region_query=aws_region_query,
            skip_ui=skip_ui,
        )

    return search_service_handler(query, ui, skip_ui)


class UI(zf.UI):
    def __init__(
        self,
        aws_region: str = "us-east-1",
        **kwargs,
    ):
        self.aws_region = aws_region
        super().__init__(
            handler=handler,
            **kwargs,
        )

    @property
    def console_domain(self) -> str:
        if self.aws_region in {"us-gov-east-1", "us-gov-west-1"}:  # pragma: no cover
            return "console.amazonaws-us-gov.com"
        else:
            return "console.aws.amazon.com"

    def format_resource_type(self, resource_type: str) -> str:
        """
        Format a resource type text with terminal color. In this project,
        the color is blue. Example:

            :blue:`sfn-statemachine`: name = CognitoUserManagement
        """
        return f"{self.terminal.green}{resource_type}{self.terminal.normal}"



# -*- coding: utf-8 -*-

from aws_console_url_search.model import (
    load_data,
)


def test_load_data():
    service_list = load_data()
    service_mapper = {srv.id: srv for srv in service_list}
    assert "s3" in service_mapper


if __name__ == "__main__":
    from aws_console_url_search.tests import run_cov_test

    run_cov_test(__file__, "aws_console_url_search.model", preview=False)



# -*- coding: utf-8 -*-

from aws_console_url_search.ui_def import (
    handler,
    UI,
)

ui = UI()


def test_search_service_handler():
    items = handler(query="ec2 inst", ui=ui, skip_ui=True)
    # items[0].enter_handler(ui)

    items = handler(query="ec2 inst!~", ui=ui, skip_ui=True)

    items = handler(query="abcdefghijklmn", ui=ui, skip_ui=True)
    assert len(items) == 1


def test_search_region_handler():
    assert ui.aws_region != "eu-west-1"
    items = handler(query="!@eu west 1", ui=ui, skip_ui=True)
    items[0].enter_handler(ui)
    assert ui.aws_region == "eu-west-1"

    items = handler(query="!@abcdefghijklmn", ui=ui, skip_ui=True)
    assert len(items) == 1


if __name__ == "__main__":
    from aws_console_url_search.tests import run_cov_test

    run_cov_test(__file__, "aws_console_url_search.ui_def", preview=False)



# -*- coding: utf-8 -*-

import zelfred.api as zf

from .ui_def import UI
from .paths import path_current_region

# ------------------------------------------------------------------------------
# We use ${HOME}/.aws_console_url_search/current_region.txt file to store
# the current region. The empty file means no region is selected.
# ------------------------------------------------------------------------------
if path_current_region.exists():
    content = path_current_region.read_text().strip()
    aws_region = content if content else None
else:
    aws_region = None

ui = UI(aws_region=aws_region)


def run_ui():
    """
    Run the AWS console url search UI. This is the entry point of the CLI command.
    """
    zf.debugger.reset()
    zf.debugger.enable()
    ui.run()



# -*- coding: utf-8 -*-

from pathlib_mate import Path

dir_python_lib = Path.dir_here(__file__)

dir_project_root = dir_python_lib.parent

# ------------------------------------------------------------------------------
# Virtual Environment Related
# ------------------------------------------------------------------------------
dir_venv = dir_project_root / ".venv"
dir_venv_bin = dir_venv / "bin"

# virtualenv executable paths
bin_pytest = dir_venv_bin / "pytest"

# test related
dir_htmlcov = dir_project_root / "htmlcov"
path_cov_index_html = dir_htmlcov / "index.html"
dir_unit_test = dir_project_root / "tests"


# ------------------------------------------------------------------------------
# ${HOME}/.aws_console_url_search/ dir related
# ------------------------------------------------------------------------------
dir_home = Path.home()
dir_aws_console_url_search = dir_home.joinpath(".aws_console_url_search")
dir_index = dir_aws_console_url_search.joinpath(".index")
dir_cache = dir_aws_console_url_search.joinpath(".cache")
path_config = dir_aws_console_url_search.joinpath("config.json")
path_current_region = dir_aws_console_url_search.joinpath("current_region.txt")

# ------------------------------------------------------------------------------
# ${dir_project_root}/aws_console_url_search/
# ------------------------------------------------------------------------------
path_data_json = dir_python_lib.joinpath("data.json")
path_region_json = dir_python_lib.joinpath("region.tsv")

dir_code = dir_python_lib.joinpath("code")
path_console_urls_json = dir_code.joinpath("console-urls.json")



# -*- coding: utf-8 -*-

from aws_console_url_search.dataset import (
    ServiceDocument,
    service_downloader,
    service_dataset,
    RegionDocument,
    region_downloader,
    region_dataset,
    preprocess_query,
)


class TestServiceDataset:
    def test_downloader(self):
        docs = service_downloader()

    def test_searcher(self):
        docs = service_dataset.search(query="ec2 inst")
        assert docs[0]["id"] == "ec2-instances"

        docs = service_dataset.search(query="s3 bucket")
        assert docs[0]["id"] == "s3-buckets"


class TestRegionDataset:
    def test_downloader(self):
        docs = region_downloader()

    def test_searcher(self):
        docs = region_dataset.search(query="us east")
        assert docs[0]["id"] == "us-east-1"


if __name__ == "__main__":
    from aws_console_url_search.tests import run_cov_test

    run_cov_test(__file__, "aws_console_url_search.dataset", preview=False)



# -*- coding: utf-8 -*-

from .model import BaseModel
from .model import Service
from .model import Menu
from .model import load_data
from .dataset import preprocess_query
from .dataset import ServiceDocument
from .dataset import service_downloader
from .dataset import service_index_name
from .dataset import service_dataset
from .dataset import RegionDocument
from .dataset import region_downloader
from .dataset import region_index_name
from .dataset import region_dataset
from .ui_def import Item
from .ui_def import InfoItem
from .ui_def import ConsoleUrlItem
from .ui_def import search_service_and_return_items
from .ui_def import search_service_handler
from .ui_def import RegionItem
from .ui_def import search_region_and_return_items
from .ui_def import search_region_handler
from .ui_def import handler
from .ui_def import UI
